!  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  RELBASE, the base routines for RELLINE and RELCONV
!  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  kernel for the XSPEC models RELLINE and RELCONV, which calculate
!  the relativsitc iron line profiles around rapidly spinning black
!  holes for positive and negative spin. The model uses a
!  precalculated table to interpolate the line profile for the desired
!  parameters.  The table was calculated by an extended version of
!  R. Speith's code (1995)
!
!  First version:  written by Thomas Dauser, 14/05/2010
!  
!  Published in the paper:
!
!  "Broad emission lines for negatively spinning black holes"
!  Dauser, Thomas; Wilms, J.; Reynolds, C. S.; Brenneman, L. W.,
!  2010, MNRAS, 409, 1534-1540  
!
! ******** current version *****************************************!
!
!  v0.4c: access to the tables via the environment variable
!         RELLINE_TABLES
!
!  v0.4b: now for h<3rg; needs a new/special table
!         (created after Dec 2012)
!  
!  v0.4a: added the relline_lp_ext and relconv_lp_ext Models to allow
!         for fitting moving, radially elongated, and accelerating
!         sources in the lamp post geometry
!
! Further infos at: www.sternwarte.uni-erlangen.de/research/relline/,
! including the explanation of the different implementations of the
! RELLINE model and the meaning of the single parameters. 
! ******************************************************************
! 
! *********** Important Notice *************************************
!  Keep in mind that all radii can be given in two units. As the
!  quantity 'radius' is positive by definition, we define that a
!  *negative radius* implies that it is given in units of the ISCO,
!  instead of gravitational radii.
!
!  I.e.:
!  r > 0: [r] = GM/c^2
!  r < 0: [r] = r_ms (r_ms = radius of marginal stability = ISCO)
!
!  For fitting data, hard and soft limits MUST be set either both 
!  positive or both negative, to ensure a reasonable outcome!
! ******************************************************************

! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MODULE par
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! defines all parameters used in variuos routines

  IMPLICIT NONE
  SAVE
  double precision :: pi=4.d0*atan(1.d0)
  integer, parameter :: nre=100, ng=20, abins=30, mu0bins=20
  ! line parameters
  double precision  :: lineE,emis1,emis2,rbreak,gamma
  double precision  :: a,theta0grad,rmin,rmax,z,rms,mu0
  integer :: limb_law
  ! energy bins given by xspec
  double precision :: emin,emax
  ! values from the table
  double precision :: trff_tab(nre,ng,2,abins,mu0bins)
  double precision :: cosne_tab(nre,ng,2,abins,mu0bins)
  double precision :: gmin_tab(nre,abins,mu0bins), & 
       gmax_tab(nre,abins,mu0bins)
  double precision :: a_tab(abins),mu0_tab(mu0bins),& 
       re(nre), re_tab(nre,abins,mu0bins),re2(nre)
  ! init table
  logical :: init_table = .false.
  logical :: init_table_jet = .false.
  logical :: warned = .false.
  logical :: warned_h1 = .false.
  logical :: warned_h2 = .false.
  logical :: warned_v = .false.
  logical :: calc = .false.
  double precision :: rb(2),delta_int
  ! interpolated values for the line calculation
  double precision :: trff(nre,ng,2), cosne(nre,ng,2), & 
       gmin(nre),gmax(nre)
  double precision :: egmin,egmax,gstar(ng),H
  real :: elo,ehi
  double precision :: intgmin, intgmax,ispec,w,rad,save_rad
  double precision :: inttrff(ng,2), intcosne(ng,2)
  ! try to save j in FUNC
  integer :: savej = 1
  ! file names
  integer :: ilo,ihi,irad
  integer, parameter :: nfrad = 1000

  character (len=240) :: & 
       relline_tables = ''


  character (len=240) :: & 
       fitstable = 'rel_table.fits'

  ! ************* for the lamp post geometry ********************** !!!!
  character (len=240) :: & 
       filename_jet = 'rel_lp_table.fits'
  ! ******** parameters for the table ***************** !!!
  integer, parameter :: n_intens =100, n_a = 20, n_h = 50
  real :: intensity(n_a,n_h,n_intens)
  real :: delta(n_a,n_h,n_intens),delta_inc(n_a,n_h,n_intens)
  real :: jet_an(n_a), jet_r(n_a,n_intens),jet_full_h(n_a,n_h)
  real :: jet_h(n_h),jet_intens(n_intens),jet_r_inter(n_intens)
  real :: jet_del(n_intens),jet_del_inc(n_intens)
  double precision :: height, fintens_jet(nfrad)

  ! ! ******** parameters for the CACHING   ***************** !!!
  real, allocatable :: cached_params_relline(:)
  real, allocatable :: cphotar_relline(:)
  real, allocatable :: cached_ear_relline(:)
  integer :: cached_ne_relline
!  logical :: no_caching_relline = .false.
  logical :: no_caching_relline = .true.
  logical :: first_cached_relline = .true.
  ! ! ******** parameters for the EXTENDED LP Routine   ***************** !!!
  double precision :: hbase,htop,vbase,v100
  integer, parameter :: h_n = 20  ! number of zones for the jet
  integer :: model_type = -1 ! 1: relline, 2: relline_lp, 3:  ! relline_lp_ext 
  integer :: cached_model_type = 0


END MODULE par



!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE relbase(EAR,NE,PARAM,PHOTAR,num_param)
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! MAIN PROGRAM
  !--------------------------------------------------
  USE par
  IMPLICIT NONE
  ! parameters required by XSPEC
  integer :: num_param
  integer :: NE
  real  :: EAR(0:NE), PHOTAR(NE)
  real :: param(num_param)
  ! parameters of the fits-table
  integer :: i,j
  double precision ::  g1, g2
  ! parameters characterizing the interpolation
  integer, parameter :: fine = 1
  real :: earfine(0:fine*NE), lspecfine(fine*NE)
  ! output parameters
  double precision :: lspec(NE), sumspec
  logical :: is_cached_relline
  !--------------------------------------------------


  ! ****************************************************
  !  *** STEP 0: check if values are already cached  ***
  ! ****************************************************

  call check_caching_relline(param,num_param,is_cached_relline,ear,ne)

  ! if true, use values from the previous iteration and return
  if (is_cached_relline) then
     do i=1,ne
        photar(i) = cphotar_relline(i)
     end do
     return
  endif
  
  ! *****************************************************
  ! * if you get here, the "model is not cached",i.e., 
  ! * normal evaluation
  ! *****************************************************

  ! *************************************
  !  *** STEP 1: load the paramters  ***
  ! *************************************
  ! initialize parameters
  call INIT_PAR(param,num_param)

  ! get the minimal and maximal values of the energy grid
  ! and write them to the PAR module
  emin = ear(0)
  emax = ear(NE)

  ! read the fits table if it isn't already loaded
  if (.not.init_table) then
     call check_table_paths()
     print *,''
     print *,'relline: initializing the precalculated table ... '
     call READ_TABLE_SINGLE
     init_table = .true.
     print *,'... done!'
     print *,''
  endif
  
  
  ! interpolate in th a-mu0-plane in order to get the correct values
  ! for trff, cosne, gmin, gmax
  call INTER_A_MU0
  


! *************************************
!  *** STEP 2: establish the grids *** 
! *************************************
  ! get the radial grid
  ! --> now done by interpolating (INTER_A_MU0)

  
  ! get the index for the smallest and largest radii
  ! and adjust rmin & rmax if they lie outside the table
  call SEARCH_RADII
  
  
  ! get the redshift-grid
  H = 2D-3
  g1 = H
  g2 = 1.D0-H
  do i=1,ng
    gstar(i) = g1 + (g2-g1)/(ng-1)*(i-1)
  enddo

  h = 2d-3
  
  ! create a finer energy grid (by factor FINE), earfine
  call FINE_GRID(EAR,NE,EARFINE,FINE,LSPECFINE)
  



  ! *************************************
  !  *** STEP 3: radial integration *** 
  ! *************************************
  call INT_RAD(lspecfine,earfine,ne,fine)

  
  ! *************************************
  !  *** STEP 4: rebin and normalize *** 
  ! *************************************
  
  sumspec = 0.D0
  do i=1,NE
     
     ! bin to the size of the input grid
     lspec(i) = 0.D0
     do j=1,FINE
       lspec(i) = lspec(i) + lspecfine((i-1)*fine + j)
     enddo
     
     ! normalize to 'cts/bin'
     lspec(i) = lspec(i)/(0.5D0*( ear(i) + ear(i-1) ))
     ! area under the line for the normalization
     sumspec = sumspec + lspec(i)
   enddo

   if (sumspec.eq.0D0) then
     sumspec = 1.D0
   endif
  ! calulate the normalized output: cts/bin
   do i=1,NE
     PHOTAR(i) = real(lspec(i)/sumspec)
  enddo

  ! cache the current evaluation if desired
  if (.not.no_caching_relline) then
     if (allocated(cphotar_relline)) then
        deallocate(cphotar_relline)
     endif 
     allocate (cphotar_relline(NE))
     do i=1,NE
        cphotar_relline(i) = photar(i)
     enddo
  endif


END SUBROUTINE relbase




!#####################################################################
!#####################################################################
! **** Routines connected to general RELBASE-code  ***************** !
!#####################################################################
!#####################################################################


! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine INT_RAD(lspecfine,earfine,ne,fine)
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! performes the radial integration
  use PAR
  implicit none
  ! array sizes
  integer :: ne,fine
  ! spingauleg integration variables
  double precision :: frad(nfrad),f_del(nfrad)
  real :: earfine(0:fine*NE), lspecfine(fine*NE)
  ! functiones called by int_rad
  double precision :: int_bin,cal_emis,gs2e
  ! integer values for radial bins
  integer :: ielo,iehi !,nfrad
  double precision :: glo,ghi,r1,r2
  ! div. variables
  integer :: i,j
!--------------------------------------------------
  
  
  ! create a finer radial grid
  r1=1.D0/rmax
  r2=1.D0/rmin

  call radial_grid(r1,r2,frad,nfrad)
  do i=1, nfrad
     frad(i) = 1/frad(i)
  enddo
  
  ! get emissivity if lamp post geometry is chosen
  if (model_type.eq.2) then
!     call inter_emis_jet(frad)
     call inter_emis_jet_beta(frad,fintens_jet,f_del,nfrad,0.d0)
  else if  (model_type.eq.3) then        
     call integrate_emis_jet_beta(frad,fintens_jet,nfrad)
  end if
  
  
  irad = ihi ! begin with the largest radius
  do i=1,nfrad
     
     ! interpolate values for the current (fine) radius
     call INTER_R(frad(i))
     
        
     ! gstar in [0,1] + corresponding energies
     glo = 0.D0
     ghi = 1.D0
     egmin = gs2e(glo)
     egmax = gs2e(ghi)
     
     ! check if the expected energy-bins are needed
     if ((egmax.gt.earfine(0)).and.(egmin.lt.earfine(ne*fine))) then
        
        
        !#2  make sure that integration is only done inside the
        !  given energy range
        if (egmin.lt.earfine(0)) then       ! lower limit
           egmin = earfine(0)
        endif
        if (egmax.gt.earfine(NE*FINE)) then ! upper limit
           egmax = earfine(NE*FINE)     
        endif
        
        !#3: search for the indices in the EARFINE-array
        !  index is such that: earfine(k-1).le.e.earfine(k)
     
        call binsrch(egmin,ielo,earfine,ne*fine+1)
        call binsrch(egmax,iehi,earfine,ne*fine+1)
        
        !#4:  PARAMETER FOR THE INTEGRATION       
        rad = frad(i)

        ! the weight of the actual integration piece is written to
        ! the module "param", because it is needed by the
        ! integration in order to make a reasonable error
        ! estimar2tion possible
  !      w = wfrad(i)

        ! TRY TRAPEZ INTEGRATION
        if (i.eq.1) then 
           w = 1d0/frad(i+1) - 1d0/frad(i)
        else if (i.eq.nfrad) then
           w = 1d0/frad(i)-1d0/frad(i-1)
        else
           w = 1d0/frad(i+1)-1d0/frad(i-1)
        endif


        ! define the emissivity law: (model 2 & 3 are lamp post)
        if (model_type.eq.1) then
           ispec = CAL_EMIS(rad)
        else
           ispec = fintens_jet(i)
        end if
        
        !#5: loor2p over the energies     
        do j=ielo,iehi
           lspecfine(j) = lspecfine(j) + & 
             real(int_bin(earfine(j-1),earfine(j),h)*w)
        enddo
     endif
  enddo                   ! r-loop

end subroutine int_rad
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function int_bin(lo,hi,h)
! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  !function which coordinates the integration over a certain energy bin
  ! input: - bin_lo and bin_hi (blo,bhi)
  !        - h (measure where approximation begins)
  ! output:  area of the bin (= luminosity = E/dt/bin) )
! -----------------------------------
  USE par, only: intgmin,intgmax,calc,delta_int,linee,save_rad,rad
  implicit none
  real :: lo,hi
  double precision :: hlo,hhi,gblo,gbhi,h
  double precision :: rlo,rhi,blo,bhi
  double precision :: lum
  double precision :: int_bin,int_edge,int_romb
  double precision :: gs2e
! -----------------------------------
  lum = 0.D0
  int_bin = 0.D0

  blo = lo
  bhi = hi

! convert bin limits to gstar
!  gblo = e2gs(blo)
!  gbhi = e2gs(bhi)

  gblo = (blo/lineE - intgmin)*delta_int!/(intgmax-intgmin)
  if (gblo.lt.0.D0) gblo = 0.D0
  if (gblo.gt.1.D0) gblo = 1.D0

  gbhi = (bhi/lineE - intgmin)*delta_int!/(intgmax-intgmin)
  if (gbhi.lt.0.D0) gbhi = 0.D0
  if (gbhi.gt.1.D0) gbhi = 1.D0


  if (gbhi.eq.0) return

! by default: bin limits are set to the integration limits
! of the 'real' integration
  rlo = blo
  rhi = bhi

! #1: lower limit approximation to be taken into account?
  if (gblo.le.h) then
     ! range of the low-app.-integration
     hlo = gblo
     hhi = h
     ! begin of the 'real' integration
     rlo = gs2e(h)
     ! .. but also check if this integration is necessary
     if (gbhi.le.h) then
        hhi = gbhi
        rlo = -1
     endif
    ! approximated integration
     lum = lum + int_edge(hlo,hhi)
  endif

  
! #2: upper limit approximation to be taken into account?
  if (gbhi.ge.(1.D0-h)) then
     ! range of the upper-app.-integration
     hhi = gbhi
     hlo = 1.D0-h
     
     ! begin of the 'real' integration
     rhi = gs2e(1-h)
     ! .. but also check if this integration is necessary
     if (gblo.ge.(1.D0-h)) then
        hlo = gblo
        rhi = -1
      endif

     ! the same approximated integration as in case #1 can be
     ! apllied, if one makes a variable transformation
     lum = lum + int_edge(hlo,hhi)
  endif


! #3: real integration (only if necessary)
  if ((rhi.ge.0).and.(rlo.ge.0)) then
     
     ! has this value benn calculated before?
     calc = .false.
     if ((rlo.eq.lo).and.(rad.eq.save_rad)) then
        calc = .true.
     endif
     
     lum = lum + int_romb(rlo,rhi)
  endif
  
  
  
! return the total luminosity in this bin
  int_bin = lum
  !return
end function int_bin
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++










! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function int_edge(blo,bhi)
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! function which makes an approximated integration for gstar->0/1
  ! this is only done within gstar=[0,H] and gstar[H,1-H]
  ! input:   bin_lo and bin_hi
  ! output:  area of the bin (= luminosity = E/dt/bin) )
! ----------------------------------- &
  USE par, only : h,intgmin,intgmax,lineE
  implicit none
  double precision :: blo,bhi,hi,lo,int_edge,hex
  double precision :: func,norm,gs2e
  integer :: k
! -----------------------------------

  norm = 0.D0

  ! get the value of the Luminosity on the point 
  ! closest to the ones to be approximated ("H")
  
!#1: upper or lower limit
!     -> write the corresponding gstar-value to hex
  if (blo.le.0.5D0) then
    hex = h
    lo = blo
    hi = bhi
  else
    hex = 1.D0-h
!    variable transformation for upper limit x -> 1 - x
!    leads to the same integral
!    (if the correct normalization is chosen; hex keeps track of that) 
    lo = 1-bhi
    hi = 1-blo
  endif
  
!#2: get the normalization value

  do k=1,2
    norm = norm + func(gs2e(hex),k)
  enddo


!#3: thanks to variable transformation:
!   both cases are described by the one for [0,H]
  norm = norm*sqrt(h)
  int_edge = 2*norm*(sqrt(hi)-sqrt(lo))*lineE*(intgmax-intgmin)


end function int_edge
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function int_romb(lo,hi)
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! function which calculates the normal integral by romberg's method
  ! (it is acutally jsut a wrapper which sets the correct parameters
  ! and coordinates the integration over k=1,2)
  ! input:   bin_lo and bin_hi
  ! output:  area of the bin (= luminosity = E/dt/bin) )
! -----------------------------------
  use par, only: linee
  implicit none
  double precision :: lo,hi,lum,int_romb
  integer :: k
  ! Romberg Integration
  double precision :: RombergIntegral, obtprec 
  double precision :: prec = 5D-2,func
  integer :: itmin = 0, itmax = 5, niter
  
! -----------------------------------
  lum = 0.D0
  

  ! Problems with the blue peak WITHOUT Romberg
  ! --> Romberg for blue peak only
  if (lo.ge.linee) then
     do k=1,2
        lum = lum + &
             RombergIntegral(lo,hi,prec,obtprec,niter,itmin,itmax,k) 
     enddo
  else
     do k=1,2
        lum = lum + FUNC((hi+lo)/2D0,k)*(hi-lo)
      enddo
  endif
  
  int_romb = lum
  return
end function int_romb
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++




! ***********************************************************************
! ***** ROMBERG INTEGRATION METHOD FOR THE RADIAL INTEGRAL **************
! ***********************************************************************


! Given function to integrate
FUNCTION FUNC(e,k)
  USE par
  implicit none
  double precision :: FUNC, e,inte,inte1
  double precision :: eg, egstar !,x,e2gs
  integer :: k,ind
  double precision :: ftrf,fmu0,limb

  ! get the redshift from the energy
  eg = e/lineE
  egstar = (eg-intgmin)*delta_int!/(intgmax-intgmin)


! find the indices in the original g-grid,
! but check first if they have already been calculated
  if ((egstar.ge.gstar(savej)).and.(egstar.lt.gstar(savej+1)))then
     ind = savej
  else
     call dbinsrch(egstar,ind,gstar,ng)     
     savej = ind
  endif

  
  
  inte = (egstar -gstar(ind))/(gstar(ind+1) - gstar(ind))
  inte1=1d0-inte
  ftrf = inte*inttrff(ind,k) + inte1*inttrff(ind+1,k)
  fmu0 = inte*intcosne(ind,k) + inte1*intcosne(ind+1,k)  

  ! isotropic limb law by default (see Svoboda (2009))
  limb = 1.D0
  if (limb_law.eq.1) then   !Laor(1991)
       limb = (1.D0 + 2.06D0*fmu0)
  else if (limb_law.eq.2) then  !Haardt (1993)
     limb = log(1.D0+1.D0/fmu0)
  endif
  
  ! INTEGRAL !
 func = (eg*rad)**3/ &
 ((intgmax-intgmin)*sqrt(egstar - egstar**2))* &
 limb*ftrf*ispec
  return 
end FUNCTION FUNC


  !*******************************************************
  !* Integral of a function FUNC(X) by Romberg's method  *
  !* --------------------------------------------------- *
  !* INPUTS:                                             *
  !*          a      begin value of x variable           *
  !*          b      end value of x variable             *
  !*       prec      desired precision                   *
  !*    itermin      minimum number of ietrations        *
  !*    itermax      maximum number of iterations        *
  !*                                                     *
  !* OUTPUTS:                                            *
  !*    obtprec      obtained precision for integral     *
  !*      niter      number of iterations done           *
  !*   integral      the integral of FUNC(X) from a to b *
  !*                                                     *
  !*******************************************************
Function RombergIntegral(a,b,prec,obtprec,niter,itermin,itermax,k)
  USE par, only: rb,calc,rad,save_rad
  implicit none
  double precision :: RombergIntegral
  integer, parameter :: MAXITER = 50
  double precision :: a,b,prec,obtprec,FUNC
  double precision :: t(0:MAXITER,0:MAXITER)
  double precision :: pas,r,s,ta,pasm
  integer :: i,j,niter,itermax,itermin,k

  if (itermax>MAXITER)  itermax=MAXITER

! EDITED HERE: check if this value has already been calculated
  if (calc) then 
     r = rb(k)
  else
     r = FUNC(a,k)
  endif

  rb(k) = FUNC(b,k)
  save_rad = rad
  
  ta = (r + rb(k)) / 2.d0
  niter=0
  ! change it to be the mean and not the integral
  pas=b-a
  pasm=1.D0
  t(0,0)=ta*pas     !!!! changed to mean !!!!
100 niter=niter+1
  pas=pas/2.d0
  pasm=pasm/2.d0
  s=ta
  do i=1,2**niter-1
     s = s + FUNC(a+pas*i,k)
  end do
  t(0,niter)=s*pas  !!!! change to mean: pasm instead of pas !!!!
  r=1.d0
  do i=1, niter
     r=r*4.d0
     j=niter-i
     t(i,j)=(r*t(i-1,j+1) - t(i-1,j))/(r-1.d0)
   end do
   ! changed routine: *1/t(niter,0) in order to compare the relative
   ! precission and not the absolute !
   obtprec = DABS(t(niter,0) - t(niter-1,0))/t(niter,0)
   if (niter > itermax)  goto 200
   if (niter<itermin)  goto 100
   if (obtprec>prec)   goto 100
200 RombergIntegral = t(niter,0)

  return
end Function RombergIntegral

! ***********************************************************************
! ***********************************************************************
! ***********************************************************************







!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE INTER_R(r)
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! this subroutine interpolates trff,cosne,gmin & gmax for 
  ! the fine radial grid
  ! *input: radius,irad
  ! *output: inttrff,intcosne,intgmin,intgmax
  USE par
  IMPLICIT NONE 
  integer :: j,k
  double precision :: intfac,r
  !---------------------------------

  if (irad.le.0) irad = ihi
   ! make sure that frad(i) lies between its 
  !two grid-neighbors, and yes: 
  !this is correct because re DEcreases
  if (.not.((r.ge.re(irad+1)).and.(r.lt.re(irad)))) then     
     irad = ihi
     do j=irad,nre
        if (r.lt.re(irad+1)) then
           irad = irad + 1
        endif
     enddo
  endif

  intfac = (r-re(irad+1))/(re(irad)-re(irad+1))
  do j=1,ng
     do k=1,2
        inttrff(j,k)=intfac*trff(irad,j,k) + &
             (1.D0-intfac)*trff(irad+1,j,k)
        intcosne(j,k)=intfac*cosne(irad,j,k) + &
             (1.D0-intfac)*cosne(irad+1,j,k)
     enddo
  enddo
  intgmin=intfac*gmin(irad)+(1.D0-intfac)*gmin(irad+1)
  intgmax=intfac*gmax(irad)+(1.D0-intfac)*gmax(irad+1)
  
  delta_int = 1.d0/(intgmax-intgmin)

END SUBROUTINE INTER_R



!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE INIT_PAR(param, num_param)
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  this routine initializes the line parameters

  !---------------------------------
  USE par
  IMPLICIT NONE
  integer :: num_param
  real :: param(num_param)
  double precision :: kerr_rms,rplus
  logical :: lp,lamp_post_or_not
  double precision :: fac = 1.1d0
  !---------------------------------

  ! Lamp Post or not??

  lp = lamp_post_or_not()

  lineE=param(1)
  
  if (lp) then    ! ***lamp post***

     ! is this the extended model ??
     if (num_param.eq.12) then
        a=param(2)
        theta0grad=param(3)
        rmin=param(4)
        rmax=param(5)
        z=param(6)
        limb_law=int(param(7))
        gamma=param(8)
        hbase=param(9)
        htop=param(10)
        vbase=param(11)
        v100=param(12)
     else 
        height=param(2)
        a=param(3)
        theta0grad=param(4)
        rmin=param(5)
        rmax=param(6)
        z=param(7)
        limb_law=int(param(8))
        gamma=param(9)
     endif


     ! for saftey: set r_break to a harmless value
     rbreak = rmax
  else            ! ***powerlaw emissivity***
     emis1=param(2)
     emis2=param(3)
     rbreak=param(4)
     a=param(5)
     theta0grad=param(6)
     rmin=param(7)
     rmax=param(8)
     z=param(9)
     limb_law=int(param(10))
  endif

  ! innermost radius is the radius of marginal stability  
  rms=kerr_rms(a)

  ! check if the UNIT is rms or rg:
  if (rmin.lt.0) rmin = -1D0*rmin*rms
  if (rmax.lt.0) rmax = -1D0*rmax*rms
  if (rbreak.lt.0) rbreak = -1D0*rbreak*rms

  ! check that rmin >=rms
  if ((rmin-rms).lt.-1d-14) then
     if (.not.warned) then
        print *,'                                          '
        print *,'Warning:         Rin < rms                '
        print *,'Change to Rin = -1, if you want to set Rin = rms'
        print *,'Rin',rmin,'rms',rms,rmin-rms
        print *,'                                          '
        warned = .true.
     endif
     rmin = rms
  endif
 
  ! check if height is in rplus or rg
  rplus = 1 + sqrt(1 - a*a)
  if (model_type.eq.2) then
     if (height.lt.0) height = -1.d0*height*rplus
     if ((height-fac*rplus).lt.-1d-8) then
        if (.not.warned_h1) then
           print *,'                                          '
           print *,'Warning (LP source too close to the BH): h < 1.1*rplus '
           print *,'Change to negtive heights (height <=',real(fac*(-1.d0)),')' 
           print *,'if you want to fit in Units of the Event Horizon, rplus'
           print *,'Height',height,'rplus',rplus
           print *,'Setting    h =  1.1*rplus  = ',rplus*fac
           print *,'                                          '
           warned_h1 = .true.
        endif
        height = rplus*fac
     endif
     
  else if (model_type.eq.3) then
     if (hbase.lt.0) hbase = -1.d0*hbase*rplus
     if (htop.lt.0) htop = -1.d0*htop*rplus
     if ((hbase-fac*rplus).lt.-1d-8) then
        if (.not.warned_h1) then
           print *,'                                          '
           print *,'Warning:         hbase < rplus                '
           print *,'Change to negtive heights (hbase <=',real(fac*(-1.d0)),')' 
           print *,'if you want to fit in Units of the Event Horizon, rplus'
           print *,'hbase',hbase,'rplus',rplus
           print *,'Setting    hbase =  1.1*rplus  = ',rplus*fac
           print *,'                                          '
           warned_h1 = .true.
        endif
        hbase = rplus*fac
     endif
     if (htop.lt.hbase) htop=hbase
  endif

  ! convert angle
  mu0=cos(theta0grad*pi/180.d0)

  ! take cosmological redshift into accout
  lineE=lineE/(1.D0+z)
  
  
END SUBROUTINE INIT_PAR



!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE READ_TABLE_SINGLE
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  !read the precalculated values from the table
  !this routine is only called, if the table isn't already loaded!
  USE par
  IMPLICIT NONE
  integer :: i,j,k
  integer :: status,unit,readwrite,blocksize
  integer :: hdunum,chdu
  integer :: hdutype,colnum
  integer :: felem=1, nulld=0
  logical :: anynull

  ! parameters to specify the opening process  
  status=0
  readwrite=0
  blocksize=1
  chdu = 1

  ! Get an unused Logical Unit Number to use to open the FITS file.
  call ftgiou(unit,status)
  
  ! Open the FITS file
  call ftopen(unit,trim(relline_tables)//'/'//trim(fitstable), & 
       readwrite,blocksize,status)
  call printfitserror(status)
  if (status.ne.0) stop

  ! check if the number of hdus is correct
  call FTTHDU(unit, hdunum, status)
  if (hdunum.ne.(abins*mu0bins +3)) then 
     print *,'Number of Extensions does not correspond to the a-mu0-grid!'
     stop
  endif


  ! get th a-grid
  chdu = chdu+1
  call FTMAHD(unit,chdu,hdutype,status)
  do i=1,abins
     colnum=1
     call ftgcvd(unit,colnum,i,felem,1,nulld,a_tab(i),anynull,status)
  enddo

  ! get the mu0grid
  chdu = chdu+1
  call FTMAHD(unit,chdu,hdutype,status)
  do i=1,mu0bins
     colnum=1
     call ftgcvd(unit,colnum,i,felem,1,nulld,mu0_tab(i),anynull,status)
  enddo


  ! loop the various a-mu0 combinations in the table for 
  ! trff,cosne, re, gmin & gmax
  ! important: first the a-loop and then the mu0-loop
  do i=1,abins
     do j=1,mu0bins
        ! move to the interessting HDU
        chdu = chdu+1
        call FTMAHD(unit,chdu,hdutype,status)
        ! read the file row by row  
        do k=1,nre
           colnum=1
           call ftgcvd(unit,colnum,k,felem,1,nulld,re_tab(k,i,j), &
                anynull,status)
           colnum=2
           call ftgcvd(unit,colnum,k,felem,1,nulld,gmin_tab(k,i,j), & 
                anynull,status)
           colnum=3
           call ftgcvd(unit,colnum,k,felem,1,nulld,gmax_tab(k,i,j), & 
                anynull,status)
           colnum=4
           call ftgcvd(unit,colnum,k,felem,ng,nulld,trff_tab(k,:,1,i,j), & 
           anynull,status)
           colnum=5
           call ftgcvd(unit,colnum,k,felem,ng,nulld,trff_tab(k,:,2,i,j), & 
                anynull,status)
           colnum=6
           call ftgcvd(unit,colnum,k,felem,ng,nulld,cosne_tab(k,:,1,i,j), & 
                anynull,status)
           colnum=7
           call ftgcvd(unit,colnum,k,felem,ng,nulld,cosne_tab(k,:,2,i,j), & 
           anynull,status)
        enddo
     enddo
  enddo
  
!  Close the file and free the unit number
  call ftclos(unit, status)
  call ftfiou(unit, status)

END SUBROUTINE READ_TABLE_SINGLE






!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE INTER_A_MU0 !(lineE,emis,a,theta0grad,rmin,rmax,z,rms)
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! interpolate trff, cosne, gmin & gmax in the a-mu0-plane
  USE par
  IMPLICIT NONE
  integer :: ia,imu0,i,j,k
  double precision :: aintfac, mu0intfac,kerr_rms,r_diff
  !    interpolations factors for mu0 and a
  ia=1

  do i=1,abins
     if (a_tab(i) .lt. a) ia=i
  enddo
  aintfac=(a-a_tab(ia))/(a_tab(ia+1)-a_tab(ia))
  
  imu0=1
  do i=1,mu0bins
     if (mu0_tab(i) .lt. mu0) imu0=i
  enddo
  mu0intfac=(mu0-mu0_tab(imu0))/(mu0_tab(imu0+1)-mu0_tab(imu0))
  
  !   if theta is too small, the smallest tabulated value is taken
  if (mu0_tab(mu0bins) .lt. mu0) then
     imu0=mu0bins-1
     mu0intfac=1
  endif
  
  !  difference from interpolating the radius linearly and from 
  !  the real value by calculating the difference at RMS
  r_diff = rms - ((1D0-aintfac)*kerr_rms(a_tab(ia)) &
       + (aintfac)*kerr_rms(a_tab(ia+1)) )

  !   interpolate now
  do i=1,nre
     re(i)=(1.0-aintfac)*re_tab(i,ia,imu0) &
          +aintfac*re_tab(i,ia+1,imu0) &
          +r_diff

     gmin(i)=(1.0-aintfac)*(1.0-mu0intfac)*gmin_tab(i,ia,imu0) &
          +aintfac*(1.0-mu0intfac)*gmin_tab(i,ia+1,imu0) & 
          +(1.0-aintfac)*mu0intfac*gmin_tab(i,ia,imu0+1) &
          +aintfac*mu0intfac*gmin_tab(i,ia+1,imu0+1)
     

     gmax(i)=(1.0-aintfac)*(1.0-mu0intfac)*gmax_tab(i,ia,imu0) &
          +aintfac*(1.0-mu0intfac)*gmax_tab(i,ia+1,imu0) &
          +(1.0-aintfac)*mu0intfac*gmax_tab(i,ia,imu0+1) & 
          +aintfac*mu0intfac*gmax_tab(i,ia+1,imu0+1)
    do j=1,ng
      do k=1,2
        trff(i,j,k)=(1.0-aintfac)*(1.0-mu0intfac) * &
        trff_tab(i,j,k,ia,imu0) &
        +aintfac*(1.0-mu0intfac)*trff_tab(i,j,k,ia+1,imu0) &
        +(1.0-aintfac)*mu0intfac*trff_tab(i,j,k,ia,imu0+1) &
        +aintfac*mu0intfac*trff_tab(i,j,k,ia+1,imu0+1) 
        
        cosne(i,j,k)=(1.0-aintfac)*(1.0-mu0intfac) * &
        cosne_tab(i,j,k,ia,imu0) &
        +aintfac*(1.0-mu0intfac)*cosne_tab(i,j,k,ia+1,imu0) &
        +(1.0-aintfac)*mu0intfac*cosne_tab(i,j,k,ia,imu0+1) &
        +aintfac*mu0intfac*cosne_tab(i,j,k,ia+1,imu0+1)
        enddo
     enddo
  enddo
  
end SUBROUTINE INTER_A_MU0





!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE SEARCH_RADII
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! subroutine for getting the smallest and the largest radius from
  ! the table
  ! if rmin and rmax lie outside of the tabulated grid they are adjusted
  ! properly
  USE par
  IMPLICIT NONE
  integer :: i

  ! default: integration over the whole AD
  ilo=nre     
  ihi=1

  ! get the indices
  do i=1,nre
     if (rmin .lt. re(i)) ilo=i
     if (rmax .lt. re(i)) ihi=i
  enddo

  ! check if r(nre) < rmin < rmax < r(1)

  if (ilo == nre) then  ! meaning: rmin < re(nre)
     rmin = re(ilo)     !  --> rmin smaller than eny tabulated value
  endif                 !   => set rmin to the smallest tabulated value
  if (ihi == 1) then
     rmax = re(ihi)     !   => set rmax to the highest tabulated value
  endif
  

END SUBROUTINE SEARCH_RADII




!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE FINE_GRID(ear,ne,earfine,fine,lspecfine) 
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! create an finer energy grid
  ! set all entries of the luminosity to 0 (important, as it is summed over)
  USE par
  IMPLICIT NONE
  integer :: ne,fine
  real :: ear(0:ne),earfine(0:ne*fine),lspecfine(ne*fine)
  integer :: i,j
  double precision :: intfac

  ! lowest bin is the same in both grids
  earfine(0) = ear(0)
  ! rest of the grid
  do i=1,ne
    do j=1,fine
        lspecfine((i-1)*fine+j)=0.d0
        intfac=float(j)/float(fine)
        earfine((i-1)*fine+j) = real(intfac)*ear(i)+ & 
             (1.0-real(intfac))*ear(i-1)
     enddo
   enddo
end SUBROUTINE FINE_GRID



!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function kerr_rms(amin)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! this functions calculates the radius of kerr_rms stability 
  !---------------------------------
  implicit none
  double precision :: kerr_rms,amin,Z1,Z2
  integer :: s
  !----------------------------------  
  ! accounts for negative spin
  S = 1
  IF (AMIN.LT.0) THEN
     S = -1
  ENDIF
  
  Z1=1.0+(1.0-amin**2.0)**(1.D0/3.D0)*((1.0+amin)**(1.D0/3.D0) &
       +(1.0-amin)**(1.D0/3.D0))
  Z2=((3.0*amin**2.0)+(Z1**2.0))**0.5
  kerr_rms=3.0+Z2-S*((3.0-Z1)*(3.0+Z1+(2*Z2)))**0.5
  return
END function kerr_rms

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function cal_emis(r)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! functions to calculate the emissivity 
  !---------------------------------
  USE par
  implicit none
  double precision :: r,cal_emis,alpha
  !----------------------------------  
  
  alpha = emis1
  if (r.ge.rbreak) then
     alpha = emis2
  endif
  cal_emis = (r/rbreak)**(-alpha)
  return
END function cal_emis




!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function e2gs(e)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! functions to calculate gstar from the energy
  !---------------------------------
  USE par, ONLY: lineE,intgmin,intgmax,delta_int
  implicit none
  double precision :: e,e2gs
  !----------------------------------  
  
  e2gs = (e/lineE - intgmin)*delta_int!/(intgmax-intgmin)
  ! check that gstar is between [0,1] and
  ! reset inappropriate values to the corresponding
  ! edge (0 or 1)

  if (e2gs.lt.0.D0) e2gs = 0.D0
  if (e2gs.gt.1.D0) e2gs = 1.D0

  return
END function e2gs



!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function gs2e(g)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! functions to calculate gstar from the energy
  !---------------------------------
  USE par, ONLY: lineE,intgmin,intgmax
  implicit none
  double precision :: g,gs2e
  !----------------------------------  
  
  gs2e = (g*(intgmax-intgmin) + intgmin)*lineE

  return
END function gs2e




!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine radial_grid(r1,r2,r,nrad)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! create a fine radial grid
  !---------------------------------
  implicit none
  integer :: nrad,i
  double precision :: r1,r2
  double precision :: r(nrad)
  
  do i=1,nrad
     r(i) = float(i-1)*(r2-r1)/nrad+r1
  enddo
  
end subroutine radial_grid


! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine printfitserror(status)
! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!    This subroutine prints out the descriptive text corresponding to the
!    error status value and prints out the contents of the internal
!    error message stack generated by FITSIO whenever an error occurs.
!     ---------------------------------
  integer :: status
  character (len=30) ::  errtext
  character (len=80) :: errmessage
!     ---------------------------------

!  Check if status is OK (no error); if so, simply return
  if (status .le. 0)return

! The FTGERR subroutine returns a descriptive 30-character text string that
!  corresponds to the integer error status number.  A complete list of all
!  the error numbers can be found in the back of the FITSIO User's Guide.
  call ftgerr(status,errtext)
  print *,'FITSIO Error Status =',status,': ',errtext

!  FITSIO usually generates an internal stack of error messages whenever
!  an error occurs.  These messages provide much more information on the
!  cause of the problem than can be provided by the single integer error
!  status value.  The FTGMSG subroutine retrieves the oldest message from
!  the stack and shifts any remaining messages on the stack down one
!  position.  FTGMSG is called repeatedly until a blank message is
!  returned, which indicates that the stack is empty.  Each error message
!  may be up to 80 characters in length.  Another subroutine, called
!  FTCMSG, is available to simply clear the whole error message stack in
!  cases where one is not interested in the contents.
  call ftgmsg(errmessage)
  do while (errmessage .ne. ' ')
    print *,errmessage
    call ftgmsg(errmessage)
  end do
end subroutine printfitserror






! ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
subroutine binsrch(e,k,ener,n)
  
  !     search for energy e in array ener(1:n) and return bin k for
  !     which ener(k).le.e.lt.ener(k+1)
  !     adapted from J. Wilms's tbvabs_new.f routine
  implicit none
  double precision :: e
  integer :: n,k,klo,khi
  real :: ener(n)
  
  klo=1
  khi=n
1 if (khi-klo.gt.1) then
     k=(khi+klo)/2
     if(ener(k).gt.e)then
        khi=k
     else
        klo=k
     endif
     goto 1
  endif
  
  if (klo.eq.0) then
     print *,'Energy out of bounds. Should not happen'
     stop
  endif
  k=klo


  if (ener(k).gt.e .or. ener(k+1).le.e ) then
     !write(*,*) ener(k),e,ener(k+1)
  endif
  
end subroutine binsrch

! ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
subroutine dbinsrch(e,k,ener,n)
  
  !     search for energy e in array ener(1:n) and return bin k for
  !     which ener(k).le.e.lt.ener(k+1)
  !     adapted from J. Wilms's tbvabs_new.f routine
  implicit none
  double precision :: e
  integer :: n,k,klo,khi
  double precision :: ener(n)
  
  klo=1
  khi=n
1 if (khi-klo.gt.1) then
     k=(khi+klo)/2
     if(ener(k).gt.e)then
        khi=k
     else
        klo=k
     endif
     goto 1
  endif
  
  if (klo.eq.0) then
     print *,'Energy out of bounds. Should not happen'
     stop
  endif
  k=klo

  
end subroutine dbinsrch


!#####################################################################
!#####################################################################
! **** Routines connected to the LAMP-POST/JET-BASE geometry ******* !
!#####################################################################
!#####################################################################



!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine check_caching_relline(param,num_param,is_cached_relline,ear,ne)
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! this routine checks if the parameters from a previous evalution 
  ! changed and tries to use previously calculated values for the current
  ! evalution to save time. These cached values are also administrated 
  ! and initialzied in this routine.

  !---------------------------------
  use par
  implicit none

  integer :: num_param,i,ne
  real :: param(num_param),ear(0:ne)
  logical :: is_cached_relline
  logical :: cache_relline = .true.
  !---------------------------------

  ! == 1 == if the variable "no_caching_relline" is set, set everything to .false.
  if(no_caching_relline) then
     is_cached_relline = .false.
     return
  end if     
  
  ! == 2 == check if the same model was evaluated before or has to be initialized
  if (model_type.eq.cached_model_type) then
     ! model was already evalated before
     cache_relline = .false.
  else
     cache_relline = .true.
     ! now we will cache the model, so lets also store this
     ! information for the next iteration
     cached_model_type = model_type
  endif


  ! ==== 3 ==== although model has been evaluted before, we need to
  !             check if anything changed
  if (.not.cache_relline) then
     ! == 3a == check if the energy grid changed 
     !          (not only number of elements, but also the values !!)
     if (size(cached_ear_relline).eq.(ne+1)) then
        do i=0,ne
           if (cached_ear_relline(i).ne.ear(i)) then
              cache_relline = .true.
           endif
        end do
     else
        cache_relline = .true.
     endif
     ! == 3b == check if any parameters changed
     if (size(cached_params_relline).eq.num_param) then
        do i=1,num_param
           if (cached_params_relline(i).ne.param(i)) then
              cache_relline = .true.
           end if
        enddo
     else
        cache_relline = .true.
     endif
     
  end if


  ! ==== 4 ==== Now we are able to decide if relline is alread cached or
  !         if no. If it's not cached, we have to do the caching, now!
  if (.not.cache_relline) then 
     ! == 4a == relline IS CACHED
     is_cached_relline = .true.
  else
     ! == 4b == relline IS NOT CACHED
     is_cached_relline = .false.

     ! == 4c == dealloacte all arrays which exist
     if (allocated(cphotar_relline)) then
        deallocate (cphotar_relline)
     endif
     if (allocated(cached_params_relline)) then
        deallocate (cached_params_relline)
     endif
     if (allocated(cached_ear_relline)) then
        deallocate (cached_ear_relline)
     endif
        

     ! == 4d == now allocate them with the correct dimension
     allocate (cached_params_relline(num_param))
     allocate (cphotar_relline(ne))
     allocate (cached_ear_relline(0:ne))
     cached_ne_relline = ne

     ! == 4d == finally cache the parameter and ear values
     do i=1,num_param
        cached_params_relline(i) = param(i)
     enddo
     do i=0,ne
        cached_ear_relline(i) = ear(i)
     enddo
     
  end if
     
  return
  
end subroutine check_caching_relline !  ++++++++++++++++++++++++++++++++++



!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function gi_potential_lp(r,a,h)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! calculates g = E/E_i in the lamp post geometry
  ! (see, e.g., page 48, Diploma Thesis, Thomas Dauser)
  !---------------------------------
  implicit none
  double precision :: gi_potential_lp
  double precision :: a,h,r,ut_h,ut_d
  !---------------------------------

  ut_d = ((r*sqrt(r)+a)/(sqrt(r)*sqrt(r**2 -3*r + 2*a*sqrt(r))))
  ut_h = sqrt((h**2 + a**2)/(h**2 - h + a**2))
  
  gi_potential_lp = ut_d/ut_h

end function gi_potential_lp  ! ++++++++++++++++++++++++++++++++++++++


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!+++++++++++++=====================+++++++++++++++++++++++++++++++++++
!++++++++++   EXTENDED JET ROUTINES  +++++++++++++++++++++++++++++++++
!+++++++++++++=====================+++++++++++++++++++++++++++++++++++
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine integrate_emis_jet_beta(frad,intens,n_rad)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  !---------------------------------
  USE PAR
  IMPLICIT NONE
  integer :: i,j,n_rad
  double precision :: frad(n_rad),intens(n_rad),arr_intens(h_n,n_rad)
  double precision :: bet,jet_speed_constant_accel,hi(0:h_n),f_del(n_rad)
  !---------------------------------

  ! #3: HEIGHT H
  ! we don't need several h-bins when hbase=htop


  if (abs(hbase-htop).lt.1e-3) then
     height = hbase
     call inter_emis_jet_beta(frad,intens,f_del,n_rad,vbase)
     return
  end if

  call create_log_hbins(hbase,htop,hi,h_n)

  do i=1,h_n
     height = 0.5*(hi(i-1)+hi(i))
     ! get the speed of the component at current height
     if (v100.gt.vbase) then
        bet = jet_speed_constant_accel(v100, height, hbase)  
     else 
        bet = vbase
        if ((.not.warned_v).and.(v100.lt.vbase)) then
           print *,' *** Warning: v100 < vbase means a decelerating jet!'
           print *,' ***          Therefore, v100 = vbase was set!'
           warned_v = .true.
        endif
     end if
     
     call inter_emis_jet_beta(frad,intens,f_del,n_rad,bet)
     ! normalize it to the height of each segment of the jet
     arr_intens(i,:) = intens*(hi(i)-hi(i-1))/(htop-hbase)
  enddo

  ! we trust/know that h_bins are linear !!!! 
  do i=1,n_rad
     intens(i) = 0.D0     

     ! this can be simply added, as we already normalized the
     ! intensity above
     do j=1,h_n
           intens(i) = arr_intens(j,i) + intens(i)
     end do     
  end do
    
end subroutine integrate_emis_jet_beta
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine inter_emis_jet_beta(frad,fintens,f_del,n_rad,bet)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! interpolate the emis-table for the lamp post geometry for the
  ! current value of the spin and height
  !---------------------------------
  USE par
  implicit none
  integer :: ka,j,kh1,kh2,k,m,i,n_rad
  real :: inter_a,inter_h(2),inter_r(2)
  double precision :: fintens_jet_a(n_a,n_rad),fintens(n_rad)
  double precision :: f_del_a(n_a,n_rad),f_del_inc_a(n_a,n_rad)
  double precision :: frad(n_rad),bet
  double precision :: gi_potential_lp_bet,doppler_factor
  double precision :: f_del(n_rad),f_del_inc(n_rad)
  real :: jet_h1(n_h),jet_h2(n_h)
  
  ! Read the table:
  if (.not.init_table_jet) then
     call READ_TABLE_JET_EXTENDED()
  end if

  ! interpolation factor in the a-plane
  call binsrch(a,ka,jet_an,n_a)
  inter_a = real((a-jet_an(ka))/(jet_an(ka+1) - jet_an(ka)))

  ! interpolation factor in the h-plane
  jet_h1 = jet_full_h(ka,:)
  jet_h2 = jet_full_h(ka+1,:)
  call binsrch(height,kh1,jet_h1,n_h)
  call binsrch(height,kh2,jet_h2,n_h)
  inter_h(1) = real((height-jet_h1(kh1))/(jet_h1(kh1+1) - jet_h1(kh1)))
  inter_h(2) = real((height-jet_h2(kh2))/(jet_h2(kh2+1) - jet_h2(kh2)))


  ! perform the interpolation 
  do j=1,n_intens
     ! #1: intensity
     jet_intens(j) =  &
          (1.0-inter_a)*(1.0-inter_h(1))*intensity(ka,kh1,j) &
          +(1.0-inter_a)*(inter_h(1))*intensity(ka,kh1+1,j) &
          +(inter_a)*(1.0-inter_h(2))*intensity(ka+1,kh2,j) &
          +(inter_a)*(inter_h(2))*intensity(ka+1,kh2+1,j)

     jet_del(j) =  &
          (1.0-inter_a)*(1.0-inter_h(1))*delta(ka,kh1,j) &
          +(1.0-inter_a)*(inter_h(1))*delta(ka,kh1+1,j) &
          +(inter_a)*(1.0-inter_h(2))*delta(ka+1,kh2,j) &
          +(inter_a)*(inter_h(2))*delta(ka+1,kh2+1,j)

     jet_del_inc(j) =  &
          (1.0-inter_a)*(1.0-inter_h(1))*delta_inc(ka,kh1,j) &
          +(1.0-inter_a)*(inter_h(1))*delta_inc(ka,kh1+1,j) &
          +(inter_a)*(1.0-inter_h(2))*delta_inc(ka+1,kh2,j) &
          +(inter_a)*(inter_h(2))*delta_inc(ka+1,kh2+1,j)


     ! #2: r-grid
     jet_r_inter(j) = &
          (1.0-inter_a)*jet_r(ka,j) +  (inter_a)*jet_r(ka+1,j)
     
  end do
 
  ! #3: bin it to the fine radial grid
  do i=1,n_rad

     do m=1,2
        call binsrch(frad(i),k,jet_r(ka+m-1,:),n_intens)

        if (jet_del(k)/pi*180.lt.75) then
           inter_r(m) = real((frad(i)-jet_r(ka+m-1,k))/ & 
                (jet_r(ka+m-1,k+1)-jet_r(ka+m-1,k)))
        else
           inter_r(m) = real((log(frad(i))-log(jet_r(ka+m-1,k)))/ & 
                (log(jet_r(ka+m-1,k+1))-log(jet_r(ka+m-1,k))))
        end if


        ! log grid ?!
        fintens_jet_a(m,i) = (1.0-inter_r(m))*log(jet_intens(k)) + &
             inter_r(m)*log(jet_intens(k+1))
        fintens_jet_a(m,i) = exp(fintens_jet_a(m,i))


        f_del_a(m,i) = (1.0-inter_r(m))*jet_del(k) + &
             inter_r(m)*jet_del(k+1)        

        f_del_inc_a(m,i) = (1.0-inter_r(m))*jet_del_inc(k) + &
             inter_r(m)*jet_del_inc(k+1)
     end do
     fintens(i) = (1.0-inter_a)*fintens_jet_a(1,i) + &
          inter_a*fintens_jet_a(2,i)
     f_del(i) = (1.0-inter_a)*f_del_a(1,i) + &
          inter_a*f_del_a(2,i)
     f_del_inc(i) = (1.0-inter_a)*f_del_inc_a(1,i) + &
          inter_a*f_del_inc_a(2,i)

     ! multiply by the additional factor gi**(gamma)
     fintens(i) = fintens(i)* & 
          gi_potential_lp_bet(frad(i),a,height,bet)**(gamma)

     ! take the beaming of the jet into account
     if (bet.gt.0.d0) then
        fintens(i) = fintens(i)*doppler_factor(f_del(i),bet)**2
     endif
  end do
  

END subroutine inter_emis_jet_beta      !+++++++++++++++++++++++++++++++


!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE READ_TABLE_JET_EXTENDED()
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  !read the precalculated values from the table
  !this routine is only called, if the table isn't already loaded!
  USE par
  IMPLICIT NONE
  integer :: i,j,k
  integer :: status,unit,readwrite,blocksize
  integer :: hdunum,chdu
  integer :: hdutype,colnum
  integer :: felem=1, nulld=0
  logical :: anynull
  real :: inp(n_intens,n_a), tab_jet_r(n_intens,n_a)
  real :: tab_jet_h(n_h,n_a)

  ! parameters to specify the opening process  
  status=0
  readwrite=0
  blocksize=1
  chdu = 1

  ! Get an unused Logical Unit Number to use to open the FITS file.
  call ftgiou(unit,status)
  
  ! Open the FITS file
  call ftopen(unit,trim(relline_tables)//'/'//trim(filename_jet), & 
       readwrite,blocksize,status)
  call printfitserror(status)
  if (status.ne.0) stop

  call FTTHDU(unit, hdunum, status)

  chdu = chdu+1
  call FTMAHD(unit,chdu,hdutype,status)
        ! read the file row by row  
  do i=1,n_a
     ! spin grid
     colnum=1
     call ftgcve(unit,colnum,i,felem,1,nulld,jet_an(i), &
          anynull,status)     
     ! r-grid
     colnum=2
     call ftgcve(unit,colnum,i,felem,n_intens,nulld,tab_jet_r(:,i), &
          anynull,status)
     do k=1,n_intens
        jet_r(i,k) = tab_jet_r(k,i)
     enddo
     ! h-grid
     colnum=3
     call ftgcve(unit,colnum,i,felem,n_h,nulld,tab_jet_h(:,i), &
          anynull,status)
     do k=1,n_h
        jet_full_h(i,k) = tab_jet_h(k,i)
     enddo



     do j=1,n_h ! intensity
        colnum = j+3
        call ftgcve(unit,colnum,i,felem,n_intens,nulld,inp(:,i), &
             anynull,status)
        do k=1,n_intens
           intensity(i,j,k) = inp(k,i)
        enddo        
     enddo

     do j=1,n_h  ! emission angle
        colnum = n_h+j+3
         call ftgcve(unit,colnum,i,felem,n_intens,nulld,inp(:,i), &
              anynull,status)
         do k=1,n_intens
            delta(i,j,k) = inp(k,i)
         enddo        
      enddo
      do j=1,n_h  ! incident angle
         colnum = 2*n_h+j+3
         call ftgcve(unit,colnum,i,felem,n_intens,nulld,inp(:,i), &
              anynull,status)
         do k=1,n_intens
            delta_inc(i,j,k) = inp(k,i)
         enddo        
      enddo

  enddo
  
!  Close the file and free the unit number
  call ftclos(unit, status)
  call ftfiou(unit, status)

END SUBROUTINE READ_TABLE_JET_EXTENDED


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
FUNCTION jet_speed_constant_accel(bet1, h, hbase)  
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  !---------------------------------
  IMPLICIT NONE
  double precision :: bet1,h,hbase,jet_speed_constant_accel
  double precision :: t,x,accel,rel_gamma
  double precision, parameter :: h1 = 100
  !---------------------------------
  
  ! bet1 is defined as the velocity at 100r_g, i.e.,
  ! we assume constant acceleration such that bet=bet1
  ! at 100r_g ABOVE the jet base hbase

  rel_gamma = 1.D0/sqrt(1-bet1**2)
  accel = 1.D0/h1* (sqrt(1 + (rel_gamma**2)*(bet1**2) ) - 1  )

  x = h-hbase
  ! time for  c = 1
  t = sqrt(x**2 + 2*x/accel)
  
  ! speed for constant accel in SRT
  jet_speed_constant_accel = accel*t / sqrt(1 + (accel**2)*(t**2))
  

END FUNCTION jet_speed_constant_accel !++++++++++++++++++++++++++++++


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine create_log_hbins(h_min,h_max,h0,hbins)
  !+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! this routine creates the spin-grid
  implicit none
  integer hbins,i
  double precision ::  h_min,h_max,h1,h2
  double precision :: h0(0:hbins)

  h1 = log(h_min)
  h2 = log(h_max)
  do i = 0,hbins
     h0(i) = ((real(i))/(hbins))
     h0(i) = h0(i)*(h2-h1)+h1
     h0(i) = exp(h0(i))
  end do

end subroutine create_log_hbins  !  ---------------------------------



! =====================================================================
function doppler_factor(del, bet)
! ---------------------------------------------------------------------
! ** Input: 
! =====================================================================
implicit none
double precision :: doppler_factor,del,bet

doppler_factor = sqrt(1.d0 - bet**2) / (1.d0 + bet*cos(del)) 

return
end function doppler_factor ! =========================================



!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function gi_potential_lp_bet(r,a,h,bet)
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! calculates g = E/E_i in the lamp post geometry
  ! (see, e.g., page 48, Diploma Thesis, Thomas Dauser)
  !---------------------------------
!  use par
  implicit none

  double precision :: gi_potential_lp_bet,bet,gam
  double precision :: a,h,r,ut_h,ut_d
  !---------------------------------
  
  gam = 1.d0/sqrt(1-bet**2)
  ut_d = ((r*sqrt(r)+a)/(sqrt(r)*sqrt(r**2 -3*r + 2*a*sqrt(r))))
  ut_h = sqrt((h**2 + a**2)/(h**2 - h + a**2))*gam
  
  gi_potential_lp_bet = ut_d/ut_h

end function gi_potential_lp_bet  ! ++++++++++++++++++++++++++++++++++++++


! ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
subroutine rev_dbinsrch(e,k,ener,n)
  
  !     search for energy e in array ener(1:n) and return bin k for
  !     which ener(k).le.e.lt.ener(k+1)
  !     adapted from J. Wilms's tbvabs_new.f routine
  !     CHANGE HERE: REVERSE SEARCH (highest value is at ener(1) !! )
  implicit none
  double precision :: e
  integer :: n,k,klo,khi
  double precision :: ener(n)
  
  klo=1
  khi=n
1 if (khi-klo.gt.1) then
     k=(khi+klo)/2
     if(ener(k).lt.e)then
        khi=k
     else
        klo=k
     endif
     goto 1
  endif
  
  if (klo.eq.0) then
     print *,'Energy out of bounds. Should not happen'
     stop
  endif
  k=klo

  
end subroutine rev_dbinsrch




function lamp_post_or_not()

  use par, only: model_type
  implicit none
  logical lamp_post_or_not

  if (model_type.eq.1) then
     lamp_post_or_not = .false.
  else if (model_type.lt.0) then
     print *,"Model_Type not defined. Should not Happen! Exiting ..."
     stop
  else 
     lamp_post_or_not = .true.
  endif

  return
end function lamp_post_or_not


subroutine check_table_paths()
  
  use par, only: relline_tables
  implicit none
  character (len=24) :: env="RELLINE_TABLES"
  character (len=240) :: var

  call GETENV(env,var)
  if (len_trim(var).gt.0) relline_tables=trim(var)
  if (len_trim(relline_tables).eq.0) relline_tables="."

end subroutine check_table_paths

